@@ -43,8 +43,8 @@ matches_uuid ()
 get_backing_device ()
 {
 	case "${1}" in
-		*.squashfs|*.ext2|*.ext3|*.ext4|*.jffs2)
-			echo $(setup_loop "${1}" "loop" "/sys/block/loop*" '0' "${LIVE_MEDIA_ENCRYPTION}" "${2}")
+		*.squashfs|*.ext2|*.ext3|*.ext4|*.jffs2|*.*.verity|*.*.fec)
+			echo $(setup_loop "${1}" '0' "${2}")
 			;;
 
 		*.dir)
@@ -95,6 +95,7 @@ is_nice_device ()
 
 check_dev ()
 {
+	local force fix
 	sysdev="${1}"
 	devname="${2}"
 	skip_uuid_check="${3}"
@@ -114,20 +115,37 @@ check_dev ()
 				ISO_DEVICE=$(dirname ${ISO_DEVICE})
 				[ -b "$ISO_DEVICE" ] && break
 				i=$(($i -1))
-		        done
+			done
 		fi
 
 		if [ "$ISO_DEVICE" = "/" ]
 		then
-			echo "Warning: device for bootoption fromiso= ($FROMISO) not found.">>/boot.log
+			# not a block device, check if it's an iso file, for
+			# example an ISO when booting on an ONIE system
+			if echo "${FROMISO}" | grep -q "\.iso$"
+			then
+				fs_type=$(get_fstype "${FROMISO}")
+				if is_supported_fs ${fs_type}
+				then
+					mkdir /run/live/fromiso
+					mount -t $fs_type "${FROMISO}" /run/live/fromiso
+					if [ "$?" != 0 ]
+					then
+						echo "Warning: unable to mount ${FROMISO}." >>/boot.log
+					fi
+					devname="/run/live/fromiso"
+				fi
+			else
+				echo "Warning: device for bootoption fromiso= ($FROMISO) not found.">>/boot.log
+			fi
 		else
 			fs_type=$(get_fstype "${ISO_DEVICE}")
 			if is_supported_fs ${fs_type}
 			then
-				mkdir /live/fromiso
-				mount -t $fs_type "$ISO_DEVICE" /live/fromiso
+				mkdir /run/live/fromiso
+				mount -t $fs_type "$ISO_DEVICE" /run/live/fromiso
 				ISO_NAME="$(echo $FROMISO | sed "s|$ISO_DEVICE||")"
-				loopdevname=$(setup_loop "/live/fromiso/${ISO_NAME}" "loop" "/sys/block/loop*" "" '')
+				loopdevname=$(setup_loop "/run/live/fromiso/${ISO_NAME}" "")
 				devname="${loopdevname}"
 			else
 				echo "Warning: unable to mount $ISO_DEVICE." >>/boot.log
@@ -185,7 +203,7 @@ check_dev ()
 
 	if [ -n "${LIVE_MEDIA_OFFSET}" ]
 	then
-		loopdevname=$(setup_loop "${devname}" "loop" "/sys/block/loop*" "${LIVE_MEDIA_OFFSET}" '')
+		loopdevname=$(setup_loop "${devname}" "${LIVE_MEDIA_OFFSET}")
 		devname="${loopdevname}"
 	fi
 
@@ -195,6 +213,33 @@ check_dev ()
 	then
 		devuid=$(blkid -o value -s UUID "$devname")
 		[ -n "$devuid" ] && grep -qs "\<$devuid\>" /var/lib/live/boot/devices-already-tried-to-mount && continue
+
+		for _PARAMETER in ${LIVE_BOOT_CMDLINE}
+		do
+			case "${_PARAMETER}" in
+				forcefsck)
+					FORCEFSCK="true"
+					;;
+			esac
+		done
+
+		if [ "${PERSISTENCE_FSCK}" = "true" ] ||  [ "${PERSISTENCE_FSCK}" = "yes" ] || [ "${FORCEFSCK}" = "true" ]
+		then
+			force=""
+			if [ "$FORCEFSCK" = "true" ]
+			then
+				force="-f"
+			fi
+
+			fix="-a"
+			if [ "$FSCKFIX" = "true" ] || [ "$FSCKFIX" = "yes" ]
+			then
+				fix="-y"
+			fi
+
+			fsck $fix $force ${devname} >> fsck.log 2>&1
+		fi
+
 		mount -t ${fstype} -o ro,noatime "${devname}" ${mountpoint} || continue
 		[ -n "$devuid" ] && echo "$devuid" >> /var/lib/live/boot/devices-already-tried-to-mount
 
@@ -203,9 +248,9 @@ check_dev ()
 			if [ -f ${mountpoint}/${FINDISO} ]
 			then
 				umount ${mountpoint}
-				mkdir -p /live/findiso
-				mount -t ${fstype} -o ro,noatime "${devname}" /live/findiso
-				loopdevname=$(setup_loop "/live/findiso/${FINDISO}" "loop" "/sys/block/loop*" 0 "")
+				mkdir -p /run/live/findiso
+				mount -t ${fstype} -o ro,noatime "${devname}" /run/live/findiso
+				loopdevname=$(setup_loop "/run/live/findiso/${FINDISO}" 0)
 				devname="${loopdevname}"
 				mount -t iso9660 -o ro,noatime "${devname}" ${mountpoint}
 			else
@@ -420,7 +465,21 @@ is_supported_fs ()
 		return 1
 	fi
 
+	# get_fstype might report "unknown" or "swap", ignore it as no such kernel module exists
+	if [ "${fstype}" = "unknown" ] || [ "${fstype}" = "swap" ]
+	then
+		return 1
+	fi
+
 	# Try to look if it is already supported by the kernel
+	# For ntfs, since user space program ntfs-3g will be used. Check ntfs-3g instead of kernel module.
+	if [ "${fstype}" = "ntfs" ]; then
+		if type ntfs-3g >/dev/null 2>&1; then
+			return 0
+		else
+			return 1
+		fi
+	fi
 	if grep -q ${fstype} /proc/filesystems
 	then
 		return 0
@@ -449,7 +508,7 @@ is_supported_fs ()
 
 get_fstype ()
 {
-	/sbin/blkid -s TYPE -o value $1 2>/dev/null
+	blkid -s TYPE -o value $1 2>/dev/null
 }
 
 where_is_mounted ()
@@ -507,7 +566,7 @@ base_path ()
 {
 	testpath="${1}"
 	mounts="$(awk '{print $2}' /proc/mounts)"
-	testpath="$(busybox realpath ${testpath})"
+	testpath="$(realpath ${testpath})"
 
 	while true
 	do
@@ -575,85 +634,44 @@ load_keymap ()
 
 setup_loop ()
 {
-	local fspath module pattern offset encryption readonly
+	local fspath module pattern offset readonly
 	fspath=${1}
-	module=${2}
-	pattern=${3}
-	offset=${4}
-	encryption=${5}
-	readonly=${6}
+	offset=${2}
+	readonly=${3}
 
 	# the output of setup_loop is evaluated in other functions,
 	# modprobe leaks kernel options like "libata.dma=0"
 	# as "options libata dma=0" on stdout, causing serious
 	# problems therefor, so instead always avoid output to stdout
-	modprobe -q -b "${module}" 1>/dev/null
+	modprobe -q -b "loop" 1>/dev/null
+	losetup -f 1>/dev/null
 
 	udevadm settle
 
-	for loopdev in ${pattern}
-	do
-		if [ "$(cat ${loopdev}/size)" -eq 0 ]
+	dev=$(losetup -f)
+	if [ "$?" != "0" ]
+	then
+		panic "No loop devices available"
+	fi
+
+	options=''
+	if [ -n "${readonly}" ]
+	then
+		if losetup --help 2>&1 | grep -q -- "-r\b"
 		then
-			dev=$(sys2dev "${loopdev}")
-			options=''
-
-			if [ -n "${readonly}" ]
-			then
-				if losetup --help 2>&1 | grep -q -- "-r\b"
-				then
-					options="${options} -r"
-				fi
-			fi
-
-			if [ -n "${offset}" ] && [ 0 -lt "${offset}" ]
-			then
-				options="${options} -o ${offset}"
-			fi
-
-			if [ -z "${encryption}" ]
-			then
-				losetup ${options} "${dev}" "${fspath}"
-			else
-				# Loop AES encryption
-				while true
-				do
-					load_keymap
-
-					echo -n "Enter passphrase for root filesystem: " >&6
-					read -s passphrase
-					echo "${passphrase}" > /tmp/passphrase
-					unset passphrase
-					exec 9</tmp/passphrase
-					/sbin/losetup ${options} -e "${encryption}" -p 9 "${dev}" "${fspath}"
-					error=${?}
-					exec 9<&-
-					rm -f /tmp/passphrase
-
-					if [ 0 -eq ${error} ]
-					then
-						unset error
-						break
-					fi
-
-					echo
-					echo -n "There was an error decrypting the root filesystem ... Retry? [Y/n] " >&6
-					read answer
-
-					if [ "$(echo "${answer}" | cut -b1 | tr A-Z a-z)" = "n" ]
-					then
-						unset answer
-						break
-					fi
-				done
-			fi
-
-			echo "${dev}"
-			return 0
+			options="${options} -r"
 		fi
-	done
+	fi
 
-	panic "No loop devices available"
+	if [ -n "${offset}" ] && [ 0 -lt "${offset}" ]
+	then
+		options="${options} -o ${offset}"
+	fi
+
+	losetup ${options} "${dev}" "${fspath}"
+
+	echo "${dev}"
+	return 0	
 }
 
 try_mount ()
@@ -697,7 +715,7 @@ mount_persistence_media ()
 
 	# get_custom_mounts() might call this with a directory path instead
 	# of a block device path. This means we have found sub-directory path
-	# underneath /lib/live/mounts/persistence, so we're done
+	# underneath /run/live/persistence, so we're done
 	if [ -d "${device}" ]
 	then
 		echo "${device}"
@@ -709,7 +727,7 @@ mount_persistence_media ()
 		return 1
 	fi
 
-	backing="/live/persistence/$(basename ${device})"
+	backing="/run/live/persistence/$(basename ${device})"
 
 	mkdir -p "${backing}"
 	old_backing="$(where_is_mounted ${device})"
@@ -721,7 +739,7 @@ mount_persistence_media ()
 		then
 			mount_opts="ro,noatime"
 		fi
-		if mount -t "${fstype}" -o "${mount_opts}" "${device}" "${backing}" >/dev/null
+		if mount -t "${fstype}" -o "${mount_opts}" "${device}" "${backing}" >/dev/null 2>&1
 		then
 			echo ${backing}
 			return 0
@@ -732,7 +750,7 @@ mount_persistence_media ()
 		fi
 	elif [ "${backing}" != "${old_backing}" ]
 	then
-		if ! mount --move ${old_backing} ${backing} >/dev/null
+		if ! mount -o move ${old_backing} ${backing} >/dev/null
 		then
 			[ -z "${probe}" ] && log_warning_msg "Failed to move persistence media ${device}"
 			rmdir "${backing}"
@@ -772,7 +790,7 @@ close_persistence_media ()
 
 	if is_active_luks_mapping ${device}
 	then
-		/sbin/cryptsetup luksClose ${device}
+		cryptsetup luksClose ${device}
 	fi
 }
 
@@ -786,7 +804,7 @@ open_luks_device ()
 		opts="${opts} --readonly"
 	fi
 
-	if /sbin/cryptsetup status "${name}" >/dev/null 2>&1
+	if cryptsetup status "${name}" >/dev/null 2>&1
 	then
 		re="^[[:space:]]*device:[[:space:]]*\([^[:space:]]*\)$"
 		opened_dev=$(cryptsetup status ${name} 2>/dev/null | grep "${re}" | sed "s|${re}|\1|")
@@ -827,7 +845,7 @@ open_luks_device ()
 	while true
 	do
 		$cryptkeyscript "$cryptkeyprompt" | \
-			/sbin/cryptsetup -T 1 luksOpen ${dev} ${name} ${opts}
+			cryptsetup -T 1 luksOpen ${dev} ${name} ${opts}
 
 		if [ 0 -eq ${?} ]
 		then
@@ -868,14 +886,14 @@ get_gpt_name ()
 {
     local dev
     dev="${1}"
-    /sbin/blkid -s PART_ENTRY_NAME -p -o value ${dev} 2>/dev/null
+    blkid -s PART_ENTRY_NAME -p -o value ${dev} 2>/dev/null
 }
 
 is_gpt_device ()
 {
     local dev
     dev="${1}"
-    [ "$(/sbin/blkid -s PART_ENTRY_SCHEME -p -o value ${dev} 2>/dev/null)" = "gpt" ]
+    [ "$(blkid -s PART_ENTRY_SCHEME -p -o value ${dev} 2>/dev/null)" = "gpt" ]
 }
 
 probe_for_gpt_name ()
@@ -915,7 +933,7 @@ probe_for_fs_label ()
 
 	for label in ${overlays}
 	do
-		if [ "$(/sbin/blkid -s LABEL -o value $dev 2>/dev/null)" = "${label}" ]
+		if [ "$(blkid -s LABEL -o value $dev 2>/dev/null)" = "${label}" ]
 		then
 			echo "${label}=${dev}"
 		fi
@@ -941,7 +959,7 @@ probe_for_file_name ()
 		if [ -f "${path}" ]
 		then
 			local loopdev
-			loopdev=$(setup_loop "${path}" "loop" "/sys/block/loop*")
+			loopdev=$(setup_loop "${path}")
 			ret="${ret} ${label}=${loopdev}"
 		fi
 	done
@@ -1020,9 +1038,9 @@ find_persistence_media ()
 	# in one union together.
 	#
 	black_listed_devices=""
-	for d in /live/rootfs/* /live/findiso /live/fromiso
+	for d in /run/live/medium /run/live/rootfs/* /run/live/findiso /run/live/fromiso
 	do
-		black_listed_devices="${black_listed_devices} $(what_is_mounted_on d)"
+		black_listed_devices="${black_listed_devices} $(what_is_mounted_on ${d})"
 	done
 
 	for dev in $(storage_devices "${black_listed_devices}" "${white_listed_devices}")
@@ -1035,7 +1053,10 @@ find_persistence_media ()
 		# in order to probe any filesystem it contains, like we do
 		# below. activate_custom_mounts() also depends on that any luks
 		# device already has been opened.
-		if is_in_comma_sep_list luks ${PERSISTENCE_ENCRYPTION} && is_luks_partition ${dev}
+		# If the option is "lukslabel" we check if the container has the
+		# label "persistence" first before opening it.
+		if is_in_comma_sep_list lukslabel ${PERSISTENCE_ENCRYPTION} && \
+		     is_luks_partition ${dev} && [ $(get_luks_label ${dev}) = "persistence" ]
 		then
 			if luks_device=$(open_luks_device "${dev}")
 			then
@@ -1044,6 +1065,16 @@ find_persistence_media ()
 				# skip $dev since we failed/chose not to open it
 				continue
 			fi
+		elif is_in_comma_sep_list luks ${PERSISTENCE_ENCRYPTION} && is_luks_partition ${dev}
+		then
+			if luks_device=$(open_luks_device "${dev}")
+			then
+				dev="${luks_device}"
+			else
+				# skip $dev since we failed/chose not to open it
+				continue
+			fi
+		# For lukslabel we check first if the container has a persistence label before opening it
 		elif ! is_in_comma_sep_list none ${PERSISTENCE_ENCRYPTION}
 		then
 			# skip $dev since we don't allow unencrypted storage
@@ -1074,18 +1105,25 @@ find_persistence_media ()
 			result=$(probe_for_file_name "${overlays}" ${dev})
 			if [ -n "${result}" ]
 			then
-			        local loopdevice
-				loopdevice=${result##*=}
-			        if is_in_comma_sep_list luks ${PERSISTENCE_ENCRYPTION} && is_luks_partition ${loopdevice}
+				local loopdevice
+				loopdevice=${result##*=}				
+				if (is_in_comma_sep_list luks ${PERSISTENCE_ENCRYPTION} || is_in_comma_sep_list lukslabel ${PERSISTENCE_ENCRYPTION}) &&  is_luks_partition ${loopdevice} 
 				then
-				        local luksfile
+					if is_in_comma_sep_list lukslabel ${PERSISTENCE_ENCRYPTION}
+					then
+						if [ $(get_luks_label ${loopdevice}) != "persistence" ]
+						then
+							continue
+						fi
+					fi
+					local luksfile
 					luksfile=""
 					if luksfile=$(open_luks_device "${loopdevice}")
 					then
-					        result=${result%%=*}
+						result=${result%%=*}
 						result="${result}=${luksfile}"
 					else
-					        losetup -d $loopdevice
+						losetup -d $loopdevice
 						result=""
 					fi
 				fi
@@ -1108,7 +1146,7 @@ find_persistence_media ()
 		# Close luks device if it isn't used
 		if [ -z "${result}" ] && [ -n "${luks_device}" ] && is_active_luks_mapping "${luks_device}"
 		then
-			/sbin/cryptsetup luksClose "${luks_device}"
+			cryptsetup luksClose "${luks_device}"
 		fi
 	done
 
@@ -1139,13 +1177,15 @@ get_mac ()
 is_luks_partition ()
 {
 	device="${1}"
-	/sbin/cryptsetup isLuks "${device}" 1>/dev/null 2>&1
+	cryptsetup isLuks "${device}" 1>/dev/null 2>&1
 }
 
 is_active_luks_mapping ()
 {
 	device="${1}"
-	/sbin/cryptsetup status "${device}" 1>/dev/null 2>&1
+	cryptsetup status "${device}" 2> /dev/null | \
+		awk '{if ($1 == "type:") print $2}' | \
+		grep 'LUKS\|LUKS2' 1>/dev/null
 }
 
 get_luks_backing_device ()
@@ -1155,6 +1195,13 @@ get_luks_backing_device ()
 		awk '{if ($1 == "device:") print $2}'
 }
 
+get_luks_label ()
+{
+	device=${1}
+	cryptsetup luksDump ${device} 2> /dev/null | \
+		awk '{if ($1 == "Label:") print $2}'
+}
+
 removable_dev ()
 {
 	output_format="${1}"
@@ -1163,6 +1210,10 @@ removable_dev ()
 
 	for sysblock in $(echo /sys/block/* | tr ' ' '\n' | grep -vE "/(loop|ram|dm-|fd)")
 	do
+		if [ ! -d "${sysblock}" ]; then
+			continue
+		fi
+
 		dev_ok=
 		if [ "$(cat ${sysblock}/removable)" = "1" ]
 		then
@@ -1208,6 +1259,10 @@ non_removable_dev ()
 
 	for sysblock in $(echo /sys/block/* | tr ' ' '\n' | grep -vE "/(loop|ram|dm-|fd)")
 	do
+		if [ ! -d "${sysblock}" ]; then
+			continue
+		fi
+
 		if [ "$(cat ${sysblock}/removable)" = "0" ]
 		then
 			case "${output_format}" in
@@ -1292,51 +1347,7 @@ do_union ()
 			rw_opt="rw"
 			ro_opt="rr+wh"
 			noxino_opt="noxino"
-			;;
 
-		unionfs-fuse)
-			rw_opt="RW"
-			ro_opt="RO"
-			;;
-
-		*)
-			rw_opt="rw"
-			ro_opt="ro"
-			;;
-	esac
-
-	case "${UNIONTYPE}" in
-		unionfs-fuse)
-			unionmountopts="-o cow -o noinitgroups -o default_permissions -o allow_other -o use_ino -o suid"
-			unionmountopts="${unionmountopts} ${unionrw}=${rw_opt}"
-			if [ -n "${unionro}" ]
-			then
-				for rofs in ${unionro}
-				do
-					unionmountopts="${unionmountopts}:${rofs}=${ro_opt}"
-				done
-			fi
-			( sysctl -w fs.file-max=391524 ; ulimit -HSn 16384
-			unionfs-fuse ${unionmountopts} "${unionmountpoint}" ) && \
-			( mkdir -p /run/sendsigs.omit.d
-			pidof unionfs-fuse >> /run/sendsigs.omit.d/unionfs-fuse || true )
-			;;
-
-		overlayfs)
-			# XXX: can multiple unionro be used? (overlayfs only handles two dirs, but perhaps they can be chained?)
-			# XXX: and can unionro be optional? i.e. can overlayfs skip lowerdir?
-			if echo ${unionro} | grep -q " "
-			then
-				panic "Multiple lower filesystems are currently not supported with overlayfs (unionro = ${unionro})."
-			elif [ -z "${unionro}"	]
-			then
-				panic "Overlayfs needs at least one lower filesystem (read-only branch)."
-			fi
-			unionmountopts="-o noatime,lowerdir=${unionro},upperdir=${unionrw}"
-			mount -t ${UNIONTYPE} ${unionmountopts} ${UNIONTYPE} "${unionmountpoint}"
-			;;
-
-		*)
 			unionmountopts="-o noatime,${noxino_opt},dirs=${unionrw}=${rw_opt}"
 			if [ -n "${unionro}" ]
 			then
@@ -1345,14 +1356,33 @@ do_union ()
 					unionmountopts="${unionmountopts}:${rofs}=${ro_opt}"
 				done
 			fi
-			mount -t ${UNIONTYPE} ${unionmountopts} ${UNIONTYPE} "${unionmountpoint}"
+			;;
+
+		overlay)
+			# XXX: can unionro be optional? i.e. can overlay skip lowerdir?
+			if [ -z "${unionro}" ]
+			then
+				panic "overlay needs at least one lower filesystem (read-only branch)."
+			fi
+			# Multiple lower layers can now be given using the the colon (":") as a
+			# separator character between the directory names.
+			unionro="$(echo ${unionro} | sed -e 's| |:|g')"
+			# overlayfs requires:
+			# + a workdir to become mounted
+			# + workdir and upperdir to reside under the same mount
+			# + workdir and upperdir to be in separate directories
+			mkdir -p "${unionrw}/rw"
+			mkdir -p "${unionrw}/work"
+			unionmountopts="-o noatime,lowerdir=${unionro},upperdir=${unionrw}/rw,workdir=${unionrw}/work,redirect_dir=on"
 			;;
 	esac
+
+	mount -t ${UNIONTYPE} ${unionmountopts} ${UNIONTYPE} "${unionmountpoint}"
 }
 
 get_custom_mounts ()
 {
-	# Side-effect: leaves $devices with persistence.conf mounted in /live/persistence
+	# Side-effect: leaves $devices with persistence.conf mounted in /run/live/persistence
 	# Side-effect: prints info to file $custom_mounts
 
 	local custom_mounts devices bindings links
@@ -1383,7 +1413,7 @@ get_custom_mounts ()
 
 		if [ -n "${LIVE_BOOT_DEBUG}" ] && [ -e "${include_list}" ]
 		then
-			cp ${include_list} /live/persistence/${persistence_list}.${device_name}
+			cp ${include_list} /run/live/persistence/${persistence_list}.${device_name}
 		fi
 
 		while read dir options # < ${include_list}
@@ -1394,9 +1424,9 @@ get_custom_mounts ()
 				continue
 			fi
 
-			if trim_path ${dir} | grep -q -e "^[^/]" -e "^/lib" -e "^/lib/live\(/.*\)\?$" -e "^/\(.*/\)\?\.\.\?\(/.*\)\?$"
+			if trim_path ${dir} | grep -q -e "^[^/]" -e "^/lib" -e "^/run/live\(/.*\)\?$" -e "^/\(.*/\)\?\.\.\?\(/.*\)\?$"
 			then
-				log_warning_msg "Skipping unsafe custom mount ${dir}: must be an absolute path containing neither the \".\" nor \"..\" special dirs, and cannot be \"/lib\", or \"/lib/live\" or any of its sub-directories."
+				log_warning_msg "Skipping unsafe custom mount ${dir}: must be an absolute path containing neither the \".\" nor \"..\" special dirs, and cannot be \"/lib\", or \"/run/live\" or any of its sub-directories."
 				continue
 			fi
 
@@ -1568,7 +1598,7 @@ activate_custom_mounts ()
 		rootfs_dest_backing=""
 		if [ -n "${opt_link}" ] || [ -n "${opt_union}" ]
 		then
-			for d in /live/rootfs/*
+			for d in /run/live/rootfs/*
 			do
 				if [ -n "${rootmnt}" ]
 				then
@@ -1586,11 +1616,11 @@ activate_custom_mounts ()
 		local cow_dir links_source
 		if [ -n "${opt_link}" ] && [ -z "${PERSISTENCE_READONLY}" ]
 		then
-			link_files ${source} ${dest} "s|^/live/|/lib/live/mount/|"
+			link_files ${source} ${dest} ""
 		elif [ -n "${opt_link}" ] && [ -n "${PERSISTENCE_READONLY}" ]
 		then
-			mkdir -p ${rootmnt}/lib/live/mount/persistence
-			links_source=$(mktemp -d ${rootmnt}/lib/live/mount/persistence/links-source-XXXXXX)
+			mkdir -p /run/live/persistence
+			links_source=$(mktemp -d /run/live/persistence/links-source-XXXXXX)
 			chown_ref ${source} ${links_source}
 			chmod_ref ${source} ${links_source}
 			# We put the cow dir in the below strange place to
@@ -1598,7 +1628,7 @@ activate_custom_mounts ()
 			# has its own directory and isn't nested with some
 			# other custom mount (if so that mount's files would
 			# be linked, causing breakage.
-			cow_dir="/live/overlay/lib/live/mount/persistence/$(basename ${links_source})"
+			cow_dir="/run/live/overlay/run/live/persistence/$(basename ${links_source})"
 			mkdir -p ${cow_dir}
 			chown_ref "${source}" "${cow_dir}"
 			chmod_ref "${source}" "${cow_dir}"
@@ -1609,13 +1639,13 @@ activate_custom_mounts ()
 			do_union ${dest} ${source} ${rootfs_dest_backing}
 		elif [ -n "${opt_bind}" ] && [ -z "${PERSISTENCE_READONLY}" ]
 		then
-			mount --bind "${source}" "${dest}"
+			mount -o bind "${source}" "${dest}"
 		elif [ -n "${opt_bind}" -o -n "${opt_union}" ] && [ -n "${PERSISTENCE_READONLY}" ]
 		then
 			# bind-mount and union mount are handled the same
 			# in read-only mode, but note that rootfs_dest_backing
 			# is non-empty (and necessary) only for unions
-			cow_dir="/live/overlay/${dest}"
+			cow_dir="/run/live/overlay/${dest}"
 			if [ -e "${cow_dir}" ] && [ -z "${opt_link}" ]
 			then
 				# If an earlier custom mount has files here
@@ -1626,6 +1656,13 @@ activate_custom_mounts ()
 			mkdir -p ${cow_dir}
 			chown_ref "${source}" "${cow_dir}"
 			chmod_ref "${source}" "${cow_dir}"
+			if [ "${UNIONTYPE}" = "overlay" ]
+			then
+				# When we use overlay we add the "/rw" postfix to our source when using it
+				# as upper layer. Therefore we also have to add it here when using it as
+				# the lower layer.
+				source="${source}/rw"
+			fi
 			do_union ${dest} ${cow_dir} ${source} ${rootfs_dest_backing}
 		fi
 
@@ -1647,3 +1684,13 @@ is_mountpoint ()
 
 	[ $(stat -fc%d:%D "${directory}") != $(stat -fc%d:%D "${directory}/..") ]
 }
+
+# Commandline is image1:roothash1,image2:roothash2 etc.
+get_dm_verity_hash ()
+{
+	local image dmverity
+	image="$1"
+	dmverity="$2"
+
+	echo "${dmverity}" | tr "," "\n" | awk -v image=${image} -F ":" '$1==image {print $2}'
+}
